"""

.. admonition:: Summary

   This example illustrates how one can change the indentation
   of an entire block of code, eliminate lines, and change the
   content much more drastically than what the previous
   examples have done.

   The idea behind this example is to help reduce the amount of typing
   required and increases readability when assigning attributes in a
   class's ``__init__()`` method.

Auto self
==========

Python is known for its concise and readable syntax. One exception
about the concisiveness is the boiler plate code that has to be
written when defining one's own class, especially if it has
many attributes, like::

    self.this_variable = this_variable
    self.that_variable = that_variable
    self.this_other_variable = this_other_variable
    self.foo = foo
    self.bar = bar

    self.baz = [] if baz is None else baz
    self.spam = bread + ham


This leads people to ask on various forums, such as
`this question on StackOverflow <https://stackoverflow.com/questions/3652851/what-is-the-best-way-to-do-automatic-attribute-assignment-in-python-and-is-it-a>`_,
how to do automatic assignment of attributes.  The answers most often given
are:

  - Don't do it; learn to live with the explicit ``self``.
  - Use a decorator, with various examples provided.


As programmers create more classes, they find the need to add their own
dunder methods, such as ``__eq__(self, other)``, ``__repr__(self)``, etc.
Eventually, they might get annoyed enough at having to re-create these methods
too often, with the occasional typo causing bugs that they jump with
joy when discovering `attrs: Classes Without Boilerplate <https://www.attrs.org/en/stable/>`_.

Starting with Python 3.7, the standard library includes
`dataclasses <https://docs.python.org/3/library/dataclasses.html>`_ which shares some
similarity with ``attrs``. However, it does require to use type hints which,
in my opinion, reduces readability; note that many programmers find that
type hints increase readability.

As a concrete example of using traditional Python notation and
dataclasses, let's consider the code given in
`PEP 557  <https://www.python.org/dev/peps/pep-0557/>`_ but reformatted with Black::

    class Application:
        def __init__(
            self,
            name,
            requirements,
            constraints=None,
            path="",
            executable_links=None,
            executables_dir=(),
        ):
            self.name = name
            self.requirements = requirements
            self.constraints = {} if constraints is None else constraints
            self.path = path
            self.executable_links = [] if executable_links is None else executable_links
            self.executables_dir = executables_dir
            self.additional_items = []

From the same PEP document, this is the proposed code
which gives the same initialization, but using the ``@dataclass``
decorator::

    from dataclasses import dataclass

    @dataclass
    class Application:
        name: str
        requirements: List[Requirement]
        constraints: Dict[str, str] = field(default_factory=dict)
        path: str = ''
        executable_links: List[str] = field(default_factory=list)
        executable_dir: Tuple[str] = ()
        additional_items: List[str] = field(init=False, default_factory=list)


This code does more than simply initializing the variables, but I
do not find it particularly readable.

So, I was wondering if it might be possible to imagine a simpler syntax.
``auto_self`` is what I came up with.

.. admonition:: That ship has sailed ...

    I realize that there is zero chance that the following syntax would
    be adopted, especially now that the ``dataclasses`` module has been added to
    the Python standard library. Still, you can try it out using
    ``auto_self`` hook.

.. code-block:: python

    class Application:
        def __init__(
            self,
            name,
            requirements,
            constraints=None,
            path="",
            executable_links=None,
            executables_dir=(),
        ):
            self .= :
                name
                requirements
                constraints = {} if __ is None else __
                path
                executable_links = [] if __ is None else __
                executables_dir
                additional_items = []

Here, I am using a new operator, ``.=``, which is meant to represent
the automatic assignment of a variable to the name that precedes
it (``self`` in this example).  I have seen this idea for such an operator before on
**python-ideas** but never for introducing a code block as I do here.

By design, any *dunder* (double underscore), ``__``, is taken to be equivalent to the variable
being initialized.  I chose a dunder instead of a single underscore ``_``
so that it could be used in a REPL without creating conflicts with the
existing use of a single underscore in Python's REPL.  I also find that it
makes it more readable.

Of course, one is not restricted to using ``self``, or having to use ``__``
everywhere. The following is completely equivalent - although I now
find it less readable, having been used to seeing ``__`` as easy to scan
placeholder::


    class Application:
        def __init__(
            cls,
            name,
            requirements,
            constraints=None,
            path="",
            executable_links=None,
            executables_dir=(),
        ):
            cls .= :
                name
                requirements
                constraints = {} if constraints is None else constraints
                path
                executable_links = [] if __ is None else executable_links
                executables_dir

            cls.additional_items = []

.. warning::

    Unlike ``@dataclass`` or ``attrs``, no additional method is
    created by ``auto_self``.

"""
from ideas import import_hook
import token_utils


def transform_source(source, **_kwargs):
    """Replaces code like::

        self .= :
            a
            b
            c = this if __ == that else ___

    by::

        self.a = a
        self.b = b
        self.c = this if c == that else c
    """
    new_tokens = []
    auto_self_block = False
    self_name = ""
    indentation = 0

    get_nb = token_utils.get_number
    get_first = token_utils.get_first
    get_first_index = token_utils.get_first_index

    for tokens in token_utils.get_lines(source):
        if auto_self_block:
            variable = get_first(tokens)
            if variable is not None:  # None would mean an empty line
                var_name = variable.string
                block_indent = variable.start_col
                if block_indent > indentation:
                    dedent = block_indent - indentation
                    if get_nb(tokens) == 1:
                        variable.string = f"{self_name}.{var_name} = {var_name}"
                        tokens = token_utils.dedent(tokens, dedent)
                    else:
                        variable.string = f"{self_name}.{var_name}"
                        for token in tokens:
                            if token.string == "__":
                                token.string = var_name
                        tokens = token_utils.dedent(tokens, dedent)
                else:
                    auto_self_block = False
        elif get_nb(tokens) == 4:
            index = get_first_index(tokens)
            if (
                tokens[index].is_identifier()
                and tokens[index + 1] == "."
                and tokens[index + 2] == "="
                and tokens[index + 1].end_col == tokens[index + 2].start_col
                and tokens[index + 3] == ":"
            ):
                self_name = tokens[index].string
                indentation = tokens[index].start_col
                auto_self_block = True
                continue
        new_tokens.extend(tokens)
    return token_utils.untokenize(new_tokens)


def add_hook(**_kwargs):
    """Creates and adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,
    )
    return hook



from ideas.examples import constants
from ideas.import_hook import remove_hook


def test_uppercase():
    hook = constants.add_hook()

    # The following module contains assertions confirming that it
    # is processed correctly when it is created.

    from . import uppercase

    print("\nAttempting to change values of uppercase attributes.")
    # The following confirm that attempts to modify it indirectly will fail
    assert uppercase.XX == 36, "Cannot change value of uppercase constant"
    del uppercase.XX
    uppercase.XX = "something else"
    assert uppercase.XX == 36, "Cannot change uppercase constant"

    # The following should not raise any error
    uppercase.new = 1
    uppercase.new = 3
    assert uppercase.new == 3, "Non constant values can be changed"
    del uppercase.new

    remove_hook(hook)


def test_final(on_prevent_change=None):
    hook = constants.add_hook(on_prevent_change=on_prevent_change)

    # The following module contains assertions confirming that it
    # is processed correctly when it is created.

    from . import final

    # The following confirm that attempts to modify constants indirectly will fail
    assert final.const == 1, "In test: confirm initial value of final constant"
    final.const = 2
    final.const += 36
    assert final.const == 1, "In test: confirm value of final constant did not change"

    # The following should not raise any error
    final.new = 1
    final.new = 3
    assert final.new == 3, "Non constant values can be changed"
    del final.new

    remove_hook(hook)


def test_final_prevent_change():
    test_final(on_prevent_change=True)




"""console.py
--------------

Adaptation of Python's console found in code.py so that it can be
used with import hooks.
"""

import ast
import platform
import os
import sys
import traceback
import tokenize

from code import InteractiveConsole

from . import __version__
from .session import config


BANNER = (
    f"Ideas Console version {__version__}. "
    + f"[Python version: {platform.python_version()}]"
)
_CONFIG = {}
CONSOLE_NAME = config.console_name


def configure(**kwargs):
    """Configures various defaults to be used by the console"""
    _CONFIG.update(kwargs)


class IdeasConsole(InteractiveConsole):
    """An interactive console that works with source transformations,
    AST transformations and Bytecode transformations.

    It should not need to be instantiated directly.
    """

    def __init__(  # pylint: disable=R0913
        self,
        source_init=None,
        transform_ast=None,
        transform_bytecode=None,
        transform_source=None,
        callback_params=None,
        console_dict=None,
        locals_=None,
    ):
        """This class builds upon Python's code.InteractiveConsole
        to work with import hooks.
        """
        self.transform_ast = transform_ast
        self.transform_bytecode = transform_bytecode
        self.transform_source = transform_source
        self.callback_params = callback_params

        # console_dict can be a custom dict-like object.
        # See ideas.examples.constants.py for an example.
        if console_dict is None:
            console_dict = {}
        if locals_ is not None:
            console_dict.update(**locals_)

        super().__init__(locals=console_dict)
        self.filename = CONSOLE_NAME

        if source_init is not None:
            try:
                exec(source_init(), self.locals)  # pylint: disable=W0122
            except Exception:  # pylint: disable=W0703
                self.showtraceback()
            else:
                print("   The following initializing code from ideas is included:\n")
                print(source_init())

    def push(self, line):
        """Push a line to the interpreter.

        The line should not have a trailing newline; it may have
        internal newlines.  The line is appended to a buffer and the
        interpreter's runsource() method is called with the
        concatenated contents of the buffer as the source.  If this
        indicates that the command was executed or invalid, the buffer
        is reset; otherwise, the command is incomplete, and the buffer
        is left as it was after the line was appended.  The return
        value is True if more input is required, False if the line was dealt
        with in some way (this is the same as runsource()).
        """
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        config.original = source

        if self.transform_source is not None:
            last_line = source.endswith("\n")  # signals the end of a block
            try:
                source = self.transform_source(
                    source, filename=CONSOLE_NAME, callback_params=self.callback_params
                )
            except tokenize.TokenError:
                # Pass on the original source so that open (, or additional)
                # can be handled correctly as either SyntaxErrors or indication
                # that more code is needed.
                more = self.runsource(source, CONSOLE_NAME)
                if not more:
                    self.resetbuffer()
                return more
            except Exception as exc:
                print("Error when executing transform_source:")
                print("   ", traceback.format_exception_only(type(exc), exc)[0].strip())
                self.resetbuffer()
                return
            # Some transformations may add some extra "\n" (usually at most one)
            if not last_line:
                source = source.rstrip("\n")
        more = self.runsource(source, CONSOLE_NAME)
        if not more:
            self.resetbuffer()
        return more

    def runsource(self, source, filename=CONSOLE_NAME, symbol="single"):
        """Compile and run some source in the interpreter.

        Arguments are as for compile_command().

        One several things can happen:

        1) The input is incorrect; compile_command() raised an
        exception (SyntaxError or OverflowError).  A syntax traceback
        will be printed .

        2) The input is incomplete, and more input is required;
        compile_command() returned None.  Nothing happens.

        3) The input is complete; compile_command() returned a code
        object.  The code is executed by calling self.runcode() (which
        also handles run-time exceptions, except for SystemExit).
        However, if an AST transformation is performed, we go back
        to the source and recompile in two steps so that we can
        perform an AST transformation.

        The return value is True in case 2, False in the other cases (unless
        an exception is raised).  The return value can be used to
        decide whether to use sys.ps1 or sys.ps2 to prompt the next
        line.
        """
        try:
            code_obj = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1
            config.print_transformed(source)
            self.showsyntaxerror(filename)
            return False

        if code_obj is None:
            # Case 2
            return True
        config.print_transformed(source)
        # Case 3

        if self.transform_ast is not None:
            # recreate code object, this time, with ast transform
            tree = ast.parse(source, filename)
            tree = self.transform_ast(tree)
            if hasattr(ast, "unparse"):
                try:
                    source = ast.unparse(tree)
                    config.print_transformed(source)
                    source += "\n"
                except RecursionError:
                    if config.show_changes:
                        print(
                            "Warning: cannot unparse the code sample to show changes."
                        )
                    code_obj = compile(tree, filename, "exec")
                except Exception as e:
                    print("Error when executing transform_ast:")
                    print("   ", traceback.format_exception_only(type(e), e)[0].strip())
                    return False
                else:
                    code_obj = self.compile(source, filename, symbol)
            else:
                code_obj = compile(tree, filename, "exec")
            if code_obj is None:
                return True

        if self.transform_bytecode is not None:
            try:
                code_obj = self.transform_bytecode(code_obj)
            except Exception as e:
                print("Error when executing transform_bytecode:")
                print("   ", traceback.format_exception_only(type(e), e)[0].strip())
                return False

        try:
            self.runcode(code_obj)
        except SystemExit:
            os._exit(0)
        return False


def start(banner=BANNER, prompt="ideas> ", locals_=None):
    """Starts a special console that works with import hooks."""
    if len(prompt) >= 4:
        sys.ps2 = (len(prompt) - 4) * " " + "... "
    sys.ps1 = "\n" + prompt
    if locals_ is None:
        locals_ = {"config": config}
    elif "Ideas" in locals_:
        if "ideas_config" in locals_:
            print("Ideas' configuration object is not available.")
        else:
            print("Ideas' configuration object is available as ideas_config")
            locals_["ideas_config"] = config
    else:
        locals_["config"] = config
    console = IdeasConsole(**_CONFIG, locals_=locals_)

    if console.transform_ast is not None and not hasattr(ast, "unparse"):
        banner += """
    AST transformations applied: you will need to explicitly
    call print() to see the result of a command since ast.unparse
    does not exist in your Python version.
    """
    console.interact(banner=banner)



from ideas.examples import implicit_multiplication

add_mul = implicit_multiplication.transform_source


def test_multiply_by_number():
    assert add_mul("2n") == "2*n", "Multiply name by 2"
    assert add_mul("3 n") == "3* n", "Multiply name by 3"
    assert add_mul("4()") == "4*()", "Multiply ( by 4"
    assert add_mul("5 ()") == "5* ()", "Multiply ( by 5"
    assert add_mul("6 7") == "6* 7", "Multiply 6 by 7"
    assert add_mul("n 9.1") == "n* 9.1", "Multiply n by 9.1"


def test_multiply_two_identifiers():
    assert add_mul("m n") == "m* n", "Multiply m by n"


def test_multiply_after_paren():
    assert add_mul("(a+b)n") == "(a+b)*n", "Multiply paren by n"
    assert add_mul("(a+b)8") == "(a+b)*8", "Multiply paren by 8"



from ideas.examples import auto_self
from ideas.import_hook import remove_hook


cls_source = """
def __init__(cls, *args, **kwargs):
    cls .= :
        a
        b
        x
        y
        c = [] if __ is None else __
        d
        f = 0 if __ < 0 else __
    cls.e = e
"""

cls_expected = """
def __init__(cls, *args, **kwargs):
    cls.a = a
    cls.b = b
    cls.x = x
    cls.y = y
    cls.c = [] if c is None else c
    cls.d = d
    cls.f = 0 if f < 0 else f
    cls.e = e
"""

pep_557_expected = """
class Application:
    def __init__(
        self,
        name,
        requirements,
        constraints=None,
        path="",
        executable_links=None,
        executables_dir=(),
    ):
        self.name = name
        self.requirements = requirements
        self.constraints = {} if constraints is None else constraints
        self.path = path
        self.executable_links = [] if executable_links is None else executable_links
        self.executables_dir = executables_dir

        self.additional_items = []
"""

pep_557_source = """
class Application:
    def __init__(
        self,
        name,
        requirements,
        constraints=None,
        path="",
        executable_links=None,
        executables_dir=(),
    ):
        self .= :
            name
            requirements
            constraints = {} if __ is None else __
            path
            executable_links = [] if __ is None else __
            executables_dir

        self.additional_items = []
"""

pep_557_source_other = """
class Application:
    def __init__(
        cls,
        name,
        requirements,
        constraints=None,
        path="",
        executable_links=None,
        executables_dir=(),
    ):
        cls .= :
            name
            requirements
            constraints = {} if constraints is None else __
            path
            executable_links = [] if __ is None else executable_links
            executables_dir

            additional_items = []
"""


def test_barebone():
    hook = auto_self.add_hook()
    result = auto_self.transform_source(cls_source)

    assert result == cls_expected, "auto_self test"

    remove_hook(hook)


def test_pep_557():
    hook = auto_self.add_hook()
    result = auto_self.transform_source(pep_557_source)

    assert result == pep_557_expected, "PEP 557 auto_self test"

    result_other = auto_self.transform_source(pep_557_source_other)
    result_other = result_other.replace("cls", "self")

    assert result_other == pep_557_expected, "PEP 557 auto_self test other"

    remove_hook(hook)


if __name__ == "__main__":
    test_barebone()
    print("Test completed successfully.")



"""
.. admonition:: Summary

    Introduces four different forms using a ``repeat`` keyword.
    It also demonstrates how to use a callback parameters.


repeat as a keyword
=======================

Let's begin with an example.

.. code-block::

    > python -m ideas -a repeat --show
    Ideas Console version 0.0.38. [Python version: 3.9.10]

    ideas> repeat 3:
       ...     print('Hello')
       ...
    ===========Transformed============
    for _9e4a6946f8d44ffca90dc9475537d39a in range( 3):
        print('Hello')

    -----------------------------
    Hello
    Hello
    Hello

    ideas>

As you can see, ``repeat n``, where ``n`` is an integer,
is converted into a for loop, with a randomly named
variable, guaranteed not to have a name used for another
object in the program.  This works well in practice.

However, suppose we wish to do some repeatable tests,
ensuring that the variable names are always the same
for all tests: we can do this using a 'callback parameter'.
This is a parameter that is given to the ``add_hook``
function used to create the import hook.
``add_hook`` must use Python dict to pass all required
callback parameters to ``create_hook``, so that they
can be passed back to the function used to transform the code.

Here's a second example using it.

.. code-block::

    >>> from ideas.examples import repeat
    >>> from ideas.console import start
    >>> from ideas.session import config
    >>> config.show_changes = True
    >>> repeat.add_hook(predictable_names=True)
    <IdeasMetaFinder object for ideas.examples.repeat>
    >>> start()
    Ideas Console version 0.0.38. [Python version: 3.9.10]

    ideas> repeat 3:
       ...     print('Hello')
       ...
    ===========Transformed============
    for _1 in range( 3):
        print('Hello')

    -----------------------------
    Hello
    Hello
    Hello

    ideas>

As you can see, the name of the for loop variable, ``_1``,
is much simpler ... and predictable.

You will need to have a look at the code for ``repeat.py`` to
fully understand how to use such callback parameters in your
own import hooks.


Adds ``repeat`` as a keyword to write loops. The four constructs supported
are::

    repeat n:
        # code

    repeat while condition:
        # code

    repeat until condition:
        # code

    repeat forever:
        # code

For example::

    repeat 3:
        a = 2
        repeat a*a:
            pass

is equivalent to::

    for unique_variable_name_1 in range(3):
        a = 2
        for unique_variable_name_2 in range(a*a):
            pass
"""
from ideas import import_hook, utils
import token_utils


class RepeatSyntaxError(Exception):
    """Currently, only raised when a repeat statement has a missing colon."""

    pass


def transform_source(source, callback_params=None, **_kwargs):
    """This function is called by the import hook loader and is used as a
    wrapper for the function where the real transformation is performed.

    It can use an optional parameter, ``callback_params``, which is
    a dict that can contain a key, ``"predictable_names"``, to indicate
    that variables created as loop counters should take a predictable form.
    """
    """Replaces instances of::

        repeat forever: -> while True:
        repeat while condition: -> while  condition:
        repeat until condition: -> while not condition:
        repeat n: -> for _uid in range(n):

    A complete repeat statement is restricted to be on a single line ending
    with a colon (optionally followed by a comment). If the colon is
    missing, a ``RepeatSyntaxError`` is raised.
    """
    if callback_params is None or "predictable_names" not in callback_params:
        predictable_names = False
    else:
        predictable_names = callback_params["predictable_names"]
    new_tokens = []
    if predictable_names:
        variable_name = utils.generate_predictable_names()
    else:
        variable_name = utils.generate_variable_names()

    for tokens in token_utils.get_lines(source):
        # a line of tokens can start with INDENT or DEDENT tokens ...
        first_token = token_utils.get_first(tokens)
        if first_token == "repeat":
            last_token = token_utils.get_last(tokens)
            if last_token != ":":
                raise RepeatSyntaxError(
                    "Missing colon for repeat statement on line "
                    + f"{first_token.start_row}\n    {first_token.line}."
                )

            repeat_index = token_utils.get_first_index(tokens)
            second_token = tokens[repeat_index + 1]
            if second_token == "forever":
                first_token.string = "while"
                second_token.string = "True"
            elif second_token == "while":
                first_token.string = "while"
                second_token.string = ""
            elif second_token == "until":
                first_token.string = "while"
                second_token.string = "not"
            else:
                first_token.string = "for %s in range(" % next(variable_name)
                last_token.string = "):"

        new_tokens.extend(tokens)

    return token_utils.untokenize(new_tokens)


def add_hook(predictable_names=False, **_kwargs):
    """Creates and adds the import hook in sys.meta_path.

    If ``predictable_names`` is set to ``True``, a callback parameter
    passed to the source transformation function will be used to
    create variable loops with predictable names."""
    callback_params = {"predictable_names": predictable_names}
    hook = import_hook.create_hook(
        transform_source=transform_source,
        callback_params=callback_params,
        hook_name=__name__,
    )
    return hook



expected = "..pp.p.p.."
result = []
for i in range(11):
    switch i:
        case in (0, 1, 4, 6, 8, 9):
            result.append(".")
        case in (2, 3, 5, 7):
            result.append("p")

result = "".join(result)
assert expected == result



"""
This import hook makes it possible to define constants, that is variables
which cannot have their initial value changed.

In this import hook, constants are identified in two ways:

1. names in all ``UPPERCASE_LETTERS``, which is a common Python convention

2. variables that were declared to be constant by the inclusion
   of a ``Final`` type hint.
"""


import types

from ideas import import_hook
from ideas.console import CONSOLE_NAME
import token_utils

shorten_path = import_hook.shorten_path
CONSTANTS = {}
DECLARED_FINAL = {}


def make_class(on_prevent_change=True):
    class ModuleWithConstants(types.ModuleType):
        """Special module type that prevents variables identified as constants
        to have their value changed.

        This is used to replace a module __class__ from the default as follows:

            module.__class__ = ModuleWithConstants

        This example only prints a message when an attempt is made to change
        the value of a constant. Alternatively, this could be logged or an
        exception could be raised.
        """

        def __setattr__(self, key, value):
            if (
                key in CONSTANTS[self.__file__]
                or key == key.upper()
                or key in DECLARED_FINAL[self.__file__]
            ):
                if on_prevent_change:
                    if callable(on_prevent_change):
                        on_prevent_change(  # noqa
                            filename=self.__file__, key=key, value=value, kind="set"
                        )
                    return
            super().__setattr__(key, value)

        def __delattr__(self, key):
            if key in CONSTANTS[self.__file__]:
                if on_prevent_change:
                    if callable(on_prevent_change):
                        on_prevent_change(  # noqa
                            filename=self.__file__, key=key, kind="delete"
                        )
                    return
            super().__delattr__(key)

    return ModuleWithConstants


class FinalDict(dict):
    """A ``FinalDict`` is a ``dict`` subclass which ensures that constants
    are not over-written.

    Constants are identified in two ways:

       1. names in all ``UPPERCASE_LETTERS``, which is a common Python convention
       2. variables that were declared to be constant by the inclusion
          of a ``Final`` type hint.

    Note: We only override methods which could result in changing the value
    of a constant.
    """

    def __init__(self, module_filename, on_prevent_change=True):
        """Initialises the instance"""
        self.__file__ = module_filename
        self.on_prevent_change = on_prevent_change
        super().__init__()

    def __setitem__(self, key, value):
        """Sets self[key] to value.

        If key is identified as a constant, it prevents changing
        its value after initial assignment.
        """
        if key in CONSTANTS[self.__file__]:
            if self.on_prevent_change:
                if callable(self.on_prevent_change):
                    self.on_prevent_change(
                        filename=self.__file__, key=key, value=value, kind="set"
                    )
                return
        if key == key.upper() or key in DECLARED_FINAL[self.__file__]:
            CONSTANTS[self.__file__][key] = value
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        """Deletes self[key] unless key is identified as a constant"""
        if key in CONSTANTS[self.__file__]:
            if self.on_prevent_change:
                if callable(self.on_prevent_change):
                    self.on_prevent_change(
                        filename=self.__file__, key=key, kind="delete"
                    )
                return
            return
        return super().__delitem__(key)

    def setdefault(self, key, default=None):
        """Insert key with a value of default if key is not in the dictionary.

        It prevents changes if the key is identified as a constant.
        """
        if key in CONSTANTS[self.__file__]:
            if self.on_prevent_change:
                if callable(self.on_prevent_change):
                    self.on_prevent_change(
                        filename=self.__file__, key=key, value=default, kind="set"
                    )
                return
        if key == key.upper() or key in DECLARED_FINAL[self.__file__]:
            CONSTANTS[self.__file__][key] = default
        return super().setdefault(key, default)

    def pop(self, key):
        """D.pop(key) -> value, remove specified key and return the corresponding value,
        unless the key is identified as a constant.

        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        if key in CONSTANTS[self.__file__]:
            if self.on_prevent_change:
                if callable(self.on_prevent_change):
                    self.on_prevent_change(
                        filename=self.__file__, key=key, kind="delete"
                    )
                return CONSTANTS[self.__file__][key]
        return super().pop(key)

    def update(self, mapping_or_iterable=(), **kwargs):
        """Updates the content of the dict from a mapping or an iterable,
        or from a list of keywords arguments.

        Keys identified as constants are prevented from changing.
        """
        if hasattr(mapping_or_iterable, "keys"):
            for key in mapping_or_iterable:
                self.__setitem__(key, mapping_or_iterable[key])
        else:
            for key, value in mapping_or_iterable:
                self.__setitem__(key, value)

        for key in kwargs:
            self.__setitem__(key, kwargs[key])


def transform_source(source, filename=None, **_kwargs):
    """Identifies simple assignments with a Final type hint, returning
    the source unchanged.

    The pattern we are looking for is::

        |python_identifier : Final ...

    where ``|`` indicates the beginning of a line.
    """
    if filename not in CONSTANTS:
        CONSTANTS[filename] = {}
    if filename not in DECLARED_FINAL:
        DECLARED_FINAL[filename] = set([])

    for tokens in token_utils.get_lines(source):
        # a line of tokens can start with DEDENT tokens ...
        if token_utils.get_number(tokens) > 3:
            index = token_utils.get_first_index(tokens)
            first_token = tokens[index]
            if (
                first_token.start_col == 0
                and first_token.is_identifier()
                and tokens[index + 1] == ":"
                and tokens[index + 2] == "Final"
            ):

                DECLARED_FINAL[filename].add(first_token.string)
    return source


def exec_(code_object, filename=None, module=None, callback_params=None, **kwargs):
    """Executes a code_object in a local instance of a ``FinalDict``.

    The argument ``globals_`` is assumed to be the original module's ``__dict__``.
    A module's ``__dict__`` is a read-only object; therefore, we cannot execute
    code in it while expecting to be able to prevent changes to variables intended
    to remain constants.  However, we can do so indirectly.

    Instead of executing code using the module's ``__dict__`` directly, we start
    by making a copy of its content into a ``FinalDict`` instance.
    We execute the code in that instance, and use its value after execution to
    update the module's ``__dict__``.
    """
    locals_ = FinalDict(
        filename, on_prevent_change=callback_params["on_prevent_change"]
    )
    locals_.update(module.__dict__)
    exec(code_object, locals_)
    module.__dict__.update(locals_)


def on_change_print(filename=None, key=None, value=None, kind=None):
    """Function called by default when an attempt is made to change the
    value of a constant.
    """
    if kind == "set":
        print(
            "You cannot change the value of %s.%s to %s"
            % (shorten_path(filename), key, value)
        )
    elif kind == "delete":
        print("You cannot delete %s in module %s." % (key, shorten_path(filename)))
    else:
        raise NotImplementedError


def add_hook(on_prevent_change=None, **_kwargs):
    """Creates and adds the import hook in sys.meta_path

    When an attempt is made to change the value of a constant, ``on_prevent_change``
    is called. By default, this function just prints a warning about what
    was done. This could be replaced by a function that logs the results or
    one that raises an exception, etc.
    """
    if on_prevent_change is None:
        on_prevent_change = on_change_print
    callback_params = {"on_prevent_change": on_prevent_change}

    module_class = make_class(**callback_params)
    console_dict = FinalDict(CONSOLE_NAME, on_prevent_change=on_prevent_change)
    CONSTANTS[CONSOLE_NAME] = {}

    hook = import_hook.create_hook(
        module_class=module_class,
        transform_source=transform_source,
        exec_=exec_,
        callback_params=callback_params,
        console_dict=console_dict,
        hook_name=__name__,
    )
    return hook



"""This module enables someone to use ``function`` as a keyword
equivalent to ``lambda``.
"""
from ideas import import_hook
import token_utils


def transform_source(source, **_kwargs):
    """This performs a simple replacement of ``function`` by ``lambda``."""
    new_tokens = []
    for token in token_utils.tokenize(source):
        # token_utils allows us to easily replace the string content
        # of any token
        if token == "function":
            token.string = "lambda"
        new_tokens.append(token)

    return token_utils.untokenize(new_tokens)


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,  # optional
    )
    return hook



"""A file containing a single class meant to keep track of various
configuration choice during a single run/session."""


class State:
    """Keeps track of various configuration parameters"""

    def __init__(self):
        self.console_name = "Ideas Console"
        self.show_original = False
        self.current_file = ""
        self.active_console = False
        self.original = ""
        self.verbose_finder = False
        self.show_changes = False

    def print_original(self, source, header="Original"):
        """Depending on configuration, can print the original source
        code of a module that was imported."""
        self.original = source
        if self.active_console:
            return
        if not self.show_original:
            return
        print(f"==========={header}============")
        print(source)
        print("-----------------------------")

    def print_transformed(self, source, header="Transformed"):
        """Depending on the configuration, can print the transformed
        output if it differs from the original source.
        """
        if not self.show_changes:
            return
        if source == self.original:
            return

        lines = source.split("\n")
        if len(lines) == 1:
            print(f"new: {lines[0]}")
            return
        if self.active_console:
            for line in lines:
                print(f"new: {line}")
        else:
            print(f"==========={header}============")
            print(source)
            print("-----------------------------")


config = State()



"""
This replaces any explicit float by a Decimal.
It can be used either as a custom codec or import hook.
"""


from ideas import custom_encoding, import_hook
import token_utils


def source_init():
    """Adds required import"""
    return "from decimal import Decimal\n"


def transform_source(source, **_kwargs):
    """Simple transformation: replaces any explicit float by a Decimal.

    By defining this function, we can also make use of Ideas' console.
    """
    tokens = token_utils.tokenize(source)
    for token in tokens:
        if token.is_number() and "." in token.string:
            token.string = f"Decimal('{token.string}')"

    return token_utils.untokenize(tokens)


def register():
    custom_encoding.register_encoding(
        encoding_name="decimal_math",
        transform_source=transform_source,
        hook_name=__name__,
    )


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        source_init=source_init,
        transform_source=transform_source,
    )
    return hook



"""Constants defined by the Final type hint"""
try:
    from typing import Final
except ImportError:

    class Final:
        pass


const: Final = 1
const = 2
const -= 2
assert const == 1, "Constant declared final changed value."

a_b = 1
a_b = 4
assert a_b == 4, "Regular variable can change its value."



"""
easy_units
----------

To be documented.


In the meantime, see
https://aroberge.blogspot.com/2022/04/natural-syntax-for-units-in-python.html
"""

from ideas import import_hook, utils
import token_utils

PREFIX = {}
PREFIX_SHOWN = False


def transform_source(
    source, callback_params=None, filename=None, module=None, **_kwargs
):
    """This function is called by the import hook loader with the named keyword
    that we specified when we created the import hook.

    It gives us the option to compare the original source and the transformed
    one. This type of additional option can be useful when debugging
    a source transformer. Furthermore, if we wish to define a source
    transformation that combines the effect of multiple existing
    transformations, we can combine the existing "inner" functions to
    create our new transformation.
    """
    global PREFIX_SHOWN
    from ideas.console import CONSOLE_NAME

    lines = source.split("\n")
    prefix = ""
    for line in lines:
        line = line.replace(" ", "").replace("\n", "")
        if line.endswith("=pint.UnitRegistry()"):
            prefix = line.split("=pint.UnitRegistry")[0] + "."
        elif line.endswith("=UnitRegistry()"):
            prefix = line.split("=UnitRegistry()")[0] + "."
        elif line.startswith("fromastropyimportunitsas"):
            prefix = line.replace("fromastropyimportunitsas", "") + "."
        elif line == "fromastropyimportunits":
            prefix = "units."

    if prefix and hasattr(module, "__name__") and module.__name__ == "__main__":
        PREFIX["main"] = prefix
        if callback_params["show_changes"]:
            utils.print_source(prefix, "Prefix")
            PREFIX_SHOWN = True
    elif prefix and filename == CONSOLE_NAME:
        PREFIX["main"] = prefix

    if not prefix and filename == CONSOLE_NAME and "main" in PREFIX:
        prefix = PREFIX["main"]

    if (
        prefix
        and not PREFIX_SHOWN
        and filename == CONSOLE_NAME
        and callback_params["show_changes"]
    ):
        utils.print_source(prefix, "Prefix")
        PREFIX_SHOWN = True

    original = source
    if callback_params["show_original"]:
        utils.print_source(source, "Original")

    source = transform_units(source, prefix)

    if callback_params["show_changes"] and original != source:
        utils.print_source(source, "New")

    return source


def transform_units(source, prefix=""):
    tokens = token_utils.tokenize(source)
    if not tokens:
        return source

    converting_units = False
    dividing_by_units = False
    prev_token = tokens[0]
    new_tokens = [prev_token]
    prev_is_number = prev_token.is_number()
    prev_is_identifier = prev_token.is_identifier()

    for token in tokens[1:]:
        # Take note of the token type before possibly changing its nature
        # by modifying its string value.
        is_identifier = token.is_identifier()
        is_number = token.is_number()

        if converting_units and prefix and is_identifier:
            token.string = prefix + token.string

        if converting_units and token == "]":
            converting_units = False
            if dividing_by_units:
                token.string = ")"
                dividing_by_units = False
            else:
                token.string = ""

        elif prev_is_number and token == "[":
            converting_units = True
            token.string = " * "

        elif converting_units and token == "/":
            dividing_by_units = True
            token.string = "/("

        elif (
            converting_units
            and is_identifier
            and (prev_is_number or prev_is_identifier)
        ):
            token.string = " * " + token.string

        elif converting_units and token == "^":
            token.string = "**"

        new_tokens.append(token)
        prev_is_number = is_number
        prev_is_identifier = is_identifier

    return token_utils.untokenize(new_tokens)


def add_hook(show_original=False, show_changes=False, **_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    callback_params = {
        "show_original": show_original,
        "show_changes": show_changes,
    }
    hook = import_hook.create_hook(
        transform_source=transform_source,
        callback_params=callback_params,
        hook_name=__name__,
    )
    return hook



from ideas.examples import function_keyword
from ideas.import_hook import remove_hook


def test_function():
    hook = function_keyword.add_hook()
    from . import my_program
    assert my_program.square(4) == 16, "The square of 4 is 16"
    remove_hook(hook)



"""
implicit_multiplication.py
---------------------------

This module is intended to demonstrate some unusual transformations
to allow someone to write equations as they would on paper
and have Python interpret them properly.
"""
from ideas import import_hook
import token_utils


def transform_source(source, **_kwargs):
    """This adds a multiplication symbol where it would be understood as
    being implicit by the normal way algebraic equations are written but would
    be a SyntaxError in Python. Thus we have::

        2n  -> 2*n
        n 2 -> n* 2
        2(a+b) -> 2*(a+b)
        (a+b)2 -> (a+b)*2
        2 3 -> 2* 3
        m n -> m* n
        (a+b)c -> (a+b)*c

    The obvious one (in algebra) being left out is something like ``n(...)``
    which is a function call - and thus valid Python syntax.
    """

    tokens = token_utils.tokenize(source)
    if not tokens:
        return tokens

    prev_token = tokens[0]
    new_tokens = [prev_token]

    for token in tokens[1:]:
        # The code has been written in a way to demonstrate that this type of
        # transformation could be done as the source is tokenized by Python.
        if (
            (
                prev_token.is_number()
                and (token.is_identifier() or token.is_number() or token == "(")
            )
            or (
                prev_token.is_identifier()
                and (token.is_identifier() or token.is_number())
            )
            or (prev_token == ")" and (token.is_identifier() or token.is_number()))
        ):
            new_tokens.append("*")
        new_tokens.append(token)
        prev_token = token

    return token_utils.untokenize(new_tokens)


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,
    )
    return hook



"""experimental_syntax_encoding.py
----------------------------------

This is just a proof of concept.

The source is assumed to be actually encoded in utf-8.
"""

import codecs
import encodings
import re

from . import console

utf8 = encodings.search_function("utf8")

TRANSFORMERS = []


def transform_source(source, **kwargs):
    """Scans a module for special construct identifying transformers
    to use. Apply each transformer in turn to the original source
    and returns a transformed source.
    """
    # Scan for special import statements
    pattern = re.compile("from experimental-syntax import (.*)")
    lines = source.split("\n")
    new_lines = []
    for line in lines:
        match = re.search(pattern, line)
        if match:
            module_name = match.group(1).strip()
            module = __import__(module_name)
            new_lines.append(getattr(module, "import_statement"))
            TRANSFORMERS.append(getattr(module, "transform_source"))
        else:
            new_lines.append(line)
    source = "\n".join(new_lines)

    for transform in TRANSFORMERS:
        source = transform(source)

    return source


def register_encoding(encoding_name=None, transform_source=None):
    def experimental_syntax_decode(input, errors="strict"):
        """Decodes the source as utf8 and use ``transform_source`` to
        perform the required source transformation.
        """
        text, length = utf8.decode(input, errors)
        text = transform_source(text)
        return text, length

    class ExperimentalSyntaxIncrementalDecoder(encodings.utf_8.IncrementalDecoder):
        def decode(self, input, final=False):
            self.buffer += input
            if final:
                buff = self.buffer
                self.buffer = b""
                text, _ = experimental_syntax_decode(buff)
                buff = text.encode("utf-8")
                return super().decode(buff, final=True)
            else:
                return ""

    def search_function(encoding):
        if encoding != encoding_name:
            return None

        return codecs.CodecInfo(
            name=encoding_name,
            encode=utf8.encode,
            decode=experimental_syntax_decode,
            incrementalencoder=utf8.incrementalencoder,
            incrementaldecoder=ExperimentalSyntaxIncrementalDecoder,
            streamreader=encodings.utf_8.StreamReader,
            streamwriter=utf8.streamwriter,
        )

    codecs.register(search_function)
    console.configure(transform_source=transform_source)


register_encoding(
    encoding_name="experimental-syntax", transform_source=transform_source
)



"""Python does 'normalization' of unicode names by default.
As a result, some different unicode names can end up representing
the same object.

This example demonstrates how different names can be prevented
from being 'normalized'.

Original idea from Sergey B. Kirpichev.
See https://github.com/aroberge/ideas/issues/13 for a reference.
"""
import io
import tokenize
import token_utils
import unicodedata
import uuid

from ideas import import_hook

_NAMES_MAP = {}


def transform_source(source, **_kwargs):
    """Transform names that would normally be 'normalized' by
    Python into different and unique variable names.
    """
    # We want to be able to identify when the source has been modified
    # in case session.config.show_changes has been set to True.
    # However, Python's tokenize module cannot untokenize faithfully:
    # some spaces might be messed up.
    # On the other hand, token_utils can untokenize perfectly, recovering
    # the original source. Unfortunately for this case, it works
    # with (already normalized) strings.
    # So, we combine the two here, just to ensure that we can
    # correctly identify later when a source has been modified.

    new_strings = []
    done_normalization = False
    g = tokenize.tokenize(io.BytesIO(source.encode()).readline)
    for token_type, token_string, _, _, _ in g:
        if token_type == tokenize.NAME:
            normalized_name = unicodedata.normalize("NFKC", token_string)
            if normalized_name != token_string:
                if token_string not in _NAMES_MAP:
                    _NAMES_MAP[token_string] = f"{normalized_name}_{uuid.uuid4().hex!s}"
                new_strings.append(_NAMES_MAP[token_string])
                done_normalization = True
            else:
                new_strings.append(token_string)
    if not done_normalization:
        return source

    new_tokens = []
    tokens = token_utils.tokenize(source)
    for token in tokens:
        if token.is_name():
            token.string = new_strings.pop(0)
        new_tokens.append(token)
    return token_utils.untokenize(new_tokens)


def new_dir(obj=None):
    """Similar to Python's dir, but shows the original name
    entered, not the transformed one.

    Note: the real Python dir() should be available as true_dir().
    """
    import inspect

    if obj is not None:
        names = dir(obj)
    else:
        names = list(inspect.currentframe().f_back.f_locals)
    for k, v in _NAMES_MAP.items():
        names = [
            name.replace(v, k)
            for name in names
            if not (name.startswith("__") and name.endswith("__"))
        ]
    return sorted(names)


true_dir = dir


def source_init():
    name = "ideas.examples.unnormalized_unicode"
    return f"""from {name} import true_dir
from {name} import new_dir as dir
"""


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    return import_hook.create_hook(
        transform_source=transform_source, source_init=source_init, hook_name=__name__
    )



"""decimal_math_with.py
------------------------

Defines a fake context environment in which explicit floats are
replaced by decimals.
"""


from ideas import import_hook
import token_utils


def source_init():
    """Adds required import"""
    return "from decimal import Decimal\n"


def transform_source(source, **_kwargs):
    """Does the following transformation::

        with float_as_Decimal:
            a = 1.0
            b = 2.0
        c = 3.0

    to::

        if True: # with float_as_Decimal:
            a = Decimal('1.0')
            b = Decimal('2.0')
        c = 3.0
    """

    new_tokens = []
    decimal_block = False

    indentation = 0
    for line in token_utils.get_lines(source):
        first = token_utils.get_first(line)
        if first is None:
            new_tokens.extend(line)
            continue
        elif first == "with":
            first_index = token_utils.get_first_index(line)
            if len(line) > first_index + 1:
                second = line[first_index + 1]
                if second == "float_as_Decimal":
                    first.string = "if"
                    second.string = "True"
                    indentation = first.start_col
                    decimal_block = True
        elif decimal_block and first.start_col > indentation:
            for token in line:
                if token.is_number() and "." in token.string:
                    token.string = f"Decimal('{token.string}')"
        else:
            indentation = first.start_col

        new_tokens.extend(line)

    return token_utils.untokenize(new_tokens)


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        source_init=source_init,
        transform_source=transform_source,
    )
    return hook



import pytest

from ideas.examples import repeat
from ideas.import_hook import remove_hook


def test_repeat():
    hook = repeat.add_hook()

    from . import my_program  # noqa
    assert my_program.one_loop == 4, "one_loop should equal 4"
    assert my_program.two_loops == 16, "two_loops should equal 16"

    remove_hook(hook)


def no_pytest_repeat_error():
    hook = repeat.add_hook()
    error_found = False
    try:
        from . import program_with_error
    except repeat.RepeatSyntaxError:
        error_found = True

    assert error_found, "RepeatSyntaxError was caught properly"
    remove_hook(hook)


def test_repeat_error():

    hook = repeat.add_hook()

    with pytest.raises(repeat.RepeatSyntaxError):
        from . import program_with_error

    remove_hook(hook)


source = """
repeat 3:
    repeat n * 22 : # comment
        pass
"""

converted = """
for _1 in range( 3):
    for _2 in range( n * 22 ): # comment
        pass
"""


def test_predictable():
    assert converted == repeat.transform_source(
        source, callback_params={'predictable_names':True}
    ), "Testing predicted conversion."


def test_predictable2():
    hook = repeat.add_hook(predictable_names=True)

    from . import my_program  # noqa
    assert my_program.one_loop == 4, "one_loop should equal 4"
    assert my_program.two_loops == 16, "two_loops should equal 16"

    remove_hook(hook)


def test_repeat_forever():
    source = "   repeat  forever:"
    assert repeat.transform_source(source) == "   while  True:", "repeat forever test"


def test_repeat_until():
    source = "repeat until condition: # a comment"
    assert repeat.transform_source(source) == "while not condition: # a comment"


def test_repeat_while():
    source = "repeat while condition:"
    # note extra space between while and condition as we removed a keyword
    assert repeat.transform_source(source) == "while  condition:"



from ideas.examples import fractions_ast
from ideas.import_hook import remove_hook


def test_simple_addition():
    hook = fractions_ast.add_hook()
    from . import addition  # noqa
    remove_hook(hook)



"""lambda_codec.py
------------------------

This codec replaces any Python identifier (token) represented by the
single Greek letter 'λ' by the corresponding string 'lambda' which is
the Python keyword.

The source is assumed to be actually encoded in utf-8.
"""


from ideas import custom_encoding
import token_utils


def transform_source(source, **_kwargs):
    """Simple transformation: replaces any single token λ by lambda.

    By defining this function, we can also make use of Ideas' console.
    """
    tokens = token_utils.tokenize(source)
    for token in tokens:
        if token == "λ":
            token.string = "lambda"
    return token_utils.untokenize(tokens)


encoding_name = "lambda_encoding"

custom_encoding.register_encoding(
    encoding_name=encoding_name,
    transform_source=transform_source,
    hook_name=__name__,
)



"""
Demonstrates how we can simply combine source transformations.
"""

from ideas import import_hook
from ideas.examples import french, repeat


additional_vocab = {
    "répéter": "repeat",
    "sansfin": "forever",
    "jusquà": "until",
}

french.fr_to_py.update(additional_vocab)


def transform_source(source, **_kwargs):
    """This function is called by the import hook loader and uses
    transformations from two other examples.
    """
    source = french.transform_source(source)
    source = repeat.transform_source(source)
    return source


def add_hook(**_kwargs):
    """Creates and adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,
        extensions=[".pyfr"],
    )
    return hook



"""Constants defined by the UPPERCASE_NAME convention"""

XX = 36
XX = 44
XX += 2
globals()["XX"] = "Sneaky"
assert XX == 36, "Uppercase constant did not change value"

a_b = 1
a_b = 4
assert a_b == 4, "Regular variable can change its value."


YY = 1, 2, 4
YY = 3, 3
YY = 1
assert YY == (1, 2, 4), "Constant tuple did not change value"



from ideas.import_hook import remove_hook
from ideas.examples import decimal_math


def test_import_with_hook():
    hook = decimal_math.add_hook()
    from . import use_hook  # noqa
    remove_hook(hook)


def test_import_with_encoding():
    decimal_math.register()
    from . import use_encoding  # noqa




# coding: lambda_encoding

square = λ x: x**2

assert square(3) == 9

print("Using lambda-encoding: λ")  # λ is not converted inside strings

if __name__ == '__main__':
    print("The square of 5 is", square(5))



"""

Combining transformations
=========================

``experimental_syntax`` is a special module that can be used to combine
transformations. It can be invoked on the command line using::

    python -m ideas -a experimental_syntax

or using something like the following::

    >>> from ideas.examples import experimental_syntax
    >>> experimental_syntax.add_hook()

From that point on, you can add transformations to be applied using
the following normally invalid Python syntax::

    from experimental-syntax import module

Here, ``module`` will be either imported from the current working directory
or from the ``ideas.examples`` subdirectory. After being imported,
the ``module`` is scanned to look for the following functions:

    - ``transform_source``
    - ``transform_ast``
    - ``transform_bytecode``
    - ``ipython_ast_node_transformer``
    - ``source_init``


"""
import ast
from importlib import import_module
import re

from ideas import import_hook

IMPORT_STATEMENT = re.compile("from experimental-syntax import (.*)")
SOURCE_TRANSFORMERS = []
AST_TRANSFORMERS = []
BYTECODE_TRANSFORMERS = []
IPYTHON_AST_NODE_TRANSFORMERS = []

try:
    IPYTHON_SHELL = get_ipython()  # noqa
except NameError:
    IPYTHON_SHELL = None


def find_module(module_name):
    """Finds and import a module based on its name."""
    # First, try to import from current working directory
    try:
        module = import_module(module_name)
    except (ImportError, ModuleNotFoundError):
        path = f"ideas.examples.{module_name}"
        try:
            module = import_module(path)
        except ImportError:
            print(f"{module_name} is not a known module.")
            return None
    return module


def add_source_transformer(module):
    """Looks for a function named ``transformed_source`` in ``module``;
    if found, adds it to the list of source transformers.
    """
    if hasattr(module, "transform_source"):
        transform = getattr(module, "transform_source")
        SOURCE_TRANSFORMERS.append(transform)


def add_source_init(module):
    """Looks for a function named ``source_init`` in ``module``;
    if found, returns its content as a string.
    """
    if hasattr(module, "source_init"):
        return getattr(module, "source_init")()
    return ""


def add_ast_transformer(module):
    """Looks for a function named ``transformed_ast`` in ``module``;
    if found, adds it to the list of AST transformers.
    """
    if hasattr(module, "transform_ast"):
        transform = getattr(module, "transform_ast")
        AST_TRANSFORMERS.append(transform)


def add_ipython_ast_node_transformer(module):
    """Determine if it is running in an IPython environment.
    If so, looks for a function named ``ipython_ast_node_transformer``
    in ``module``; if found, adds it to the list of AST transformers
    for IPython.
    """
    if IPYTHON_SHELL is None:
        return
    if hasattr(module, "ipython_ast_node_transformer"):
        transform = getattr(module, "ipython_ast_node_transformer")
        IPYTHON_SHELL.ast_transformers.append(transform)


def add_bytecode_transformer(module):
    """Looks for a function named ``transformed_bytecode`` in ``module``;
    if found, adds it to the list of bytecode transformers.
    """
    if hasattr(module, "transform_bytecode"):
        transform = getattr(module, "transform_bytecode")
        BYTECODE_TRANSFORMERS.append(transform)


def identify_experimental_import_statements(source):
    """Identifies if a special import statement of the form

        from experimental-syntax import some_module

    If so, imports the relevant transformers found in that module.
    """
    lines = source.split("\n")
    new_lines = []
    for line in lines:
        match = re.search(IMPORT_STATEMENT, line)
        if match:
            # Remove the invalid syntax statement
            new_lines.append("\n")
            # and scan the module to find the relevant functions.
            module_name = match.group(1).strip()
            module = find_module(module_name)
            if module is None:
                continue
            add_source_transformer(module)
            source_init = add_source_init(module)
            for init_line in source_init.splitlines():
                new_lines.append(init_line)
            add_ast_transformer(module)
            add_ipython_ast_node_transformer(module)
            add_bytecode_transformer(module)
        else:
            new_lines.append(line)
    return "\n".join(new_lines)


def transform_source(source, **_kwargs):
    """Applies source transformations."""
    source = identify_experimental_import_statements(source)
    for transform in SOURCE_TRANSFORMERS:
        source = transform(source)

    return source


def transform_ast(tree, **_kwargs):
    """Transforms the Abstract Syntax Tree"""
    for transform in AST_TRANSFORMERS:
        tree = transform(tree)
        ast.fix_missing_locations(tree)
    return tree


def ipython_ast_node_transformer(node):
    """Transform an AST node when running in an IPython environment."""
    for transform in IPYTHON_AST_NODE_TRANSFORMERS:
        try:
            node = transform(node)
        except Exception:  # pylint: disable=W0703
            pass
    return node


def transform_bytecode(byte_code):
    """Applies bytecode transformations."""
    for transform in BYTECODE_TRANSFORMERS:
        byte_code = transform(byte_code)
    return byte_code


def add_hook(**_kwargs):
    """Creates and adds the import hook in sys.meta_path.
    Uses a custom extension for the exception hook."""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        transform_ast=transform_ast,
        transform_bytecode=transform_bytecode,
        hook_name=__name__,
    )
    return hook



"""Enabling emojis as normal characters.

Adapted from https://github.com/gahjelle/pythonji
MIT license; author: Geir Arne Hjelle
"""
import ast
import sys

try:
    import emoji  # noqa
except ModuleNotFoundError:
    print("You need to install 'emoji' from pypi.")
    sys.exit()

from ideas import import_hook


DELIMITERS = ("__pythonji__", "__")


def transform_source(source, **_kwargs):
    """Transform emojis into valid strings for Python"""
    return emoji.demojize(source, delimiters=DELIMITERS)


class EmojiTransformer(ast.NodeTransformer):
    """Transform strings intended to represent emojis back into emojis"""

    def visit_Str(self, node):
        return ast.copy_location(
            ast.Str(s=emoji.emojize(node.s, delimiters=DELIMITERS)), node
        )


def transform_ast(tree, **_kwargs):
    """Transforms the Abstract Syntax Tree or a single node"""
    tree_or_node = EmojiTransformer().visit(tree)
    ast.fix_missing_locations(tree_or_node)
    return tree_or_node


def add_hook(**_kwargs):
    """Creates and adds the import hook in sys.meta_path.
    Uses a custom extension for the exception hook."""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        transform_ast=transform_ast,
        hook_name=__name__,
        extensions=[".🐍"],
    )
    return hook



"""switch.py
-----------------

Implements something similar to version 1.B of
`PEP 3103 <https://www.python.org/dev/peps/pep-3103>`_
"""
from ideas import import_hook, utils
import token_utils


def transform_source(source, callback_params=None, **_kwargs):
    """Replaces code like::

        switch EXPR:
            case EXPR_1:
                SUITE
            case EXPR_2:
                SUITE
            case in EXPR_3, EXPR_4, ...:
                SUITE
            ...
            else:
                SUITE

    by::

        var_name = EXPR
        if var_name == EXPR_1:
                SUITE
        elif var_name == EXPR_2:
                SUITE
        elif var_name in EXPR_3, EXPR_4, ...:
                SUITE
        else:
                SUITE
        del var_name

    Limitation: switch blocks cannot be part of a SUITE of another switch block.
    """
    if callback_params is None or "predictable_names" not in callback_params:
        predictable_names = False
    else:
        predictable_names = callback_params["predictable_names"]
    new_tokens = []
    switch_block = False
    first_case = False
    if predictable_names:
        variable_name = utils.generate_predictable_names()
    else:
        variable_name = utils.generate_variable_names()

    for line in token_utils.get_lines(source):
        first_token = token_utils.get_first(line)
        if first_token is None:
            new_tokens.extend(line)
            continue

        if len(line) > 1:
            _index = token_utils.get_first_index(line)
            second_token = line[_index + 1]
        else:
            second_token = None

        if not switch_block:
            if first_token == "switch":
                switch_indent = first_token.start_col
                var_name = next(variable_name)
                first_token.string = f"{var_name} ="
                switch_block = True
                first_case = True
                colon = token_utils.get_last(line)
                colon.string = ""
        else:
            if first_token.start_col == switch_indent:
                switch_block = False
                new_tokens.extend([" " * switch_indent + f"del {var_name}\n"])

            elif first_token == "case" or first_token == "else":
                if first_case and first_token == "case":
                    if second_token == "in":
                        first_token.string = f"if {var_name}"
                    else:
                        first_token.string = f"if {var_name} =="
                    first_case = False
                elif first_token == "case":
                    if second_token == "in":
                        first_token.string = f"elif {var_name}"
                    else:
                        first_token.string = f"elif {var_name} =="
                dedent = first_token.start_col - switch_indent
                line = token_utils.dedent(line, dedent)

        new_tokens.extend(line)
    return token_utils.untokenize(new_tokens)


def add_hook(
    predictable_names=False,
    **_kwargs,
):
    """Creates and adds the import hook in sys.meta_path"""
    callback_params = {"predictable_names": predictable_names}
    hook = import_hook.create_hook(
        transform_source=transform_source,
        callback_params=callback_params,
        hook_name=__name__,
    )
    return hook



"""
Fractional math (token based)
=============================

.. admonition:: Summary

   This is includes a source transformation that converts
   integers literals that are followed by the ``/`` operator
   into instances of the Fraction class in the source using the tokenizer.

   In a later section, we do something similar but using an
   Abstract Syntax Tree (AST) transformation, which is a more robust
   approach for this type of example.

Consider the following standard Python code::

    >>> x = 1/10
    >>> for i in range(11):
    ...    print(i * x)
    ...
    0.0
    0.1
    0.2
    0.30000000000000004
    0.4
    0.5
    0.6000000000000001
    0.7000000000000001
    0.8
    0.9
    1.0

This is quite surprising for beginners, not familiar with the
limitations of representing floating point numbers.

However, we can "fix" this using an import hook that wraps
integer, which are followed by a ``/``, into a ``Fraction`` instance::

    > python -m ideas -a fractions_tok
       The following initializing code from ideas is included:

    from fractions import Fraction

    Ideas Console version 0.0.38. [Python version: 3.7.9]

    ideas> x =  1 / 10

    ideas> for i in range(11):
      ...     print(i * x)
      ...
    0
    1/10
    1/5
    3/10
    2/5
    1/2
    3/5
    7/10
    4/5
    9/10
    1

    ideas> from ideas.session import config

    ideas> config.show_changes = True

    ideas> x = 1 / 10
    new: x = Fraction(1) / 10

    ideas>

This example was created after a similar example using AST transformation
was created as a proof of concept. For more details about the
difference, please have a look at the AST-based example.
"""
from ideas import import_hook
import token_utils


def transform_source(source, **_kwargs):
    """Replace integers (followed by /) by Fraction objects"""
    tokens = token_utils.tokenize(source)
    if len(tokens) < 3:
        return source

    new_tokens = []
    for token, next_ in zip(tokens, tokens[1:]):
        if token.is_integer() and next_ == "/":
            token.string = f"Fraction({token.string})"
        new_tokens.append(token)
    new_tokens.append(next_)  # noqa

    return token_utils.untokenize(new_tokens)


def source_init():
    """Adds required import so that ``Fraction`` is a known object."""
    import_fraction = "from fractions import Fraction\n"
    return import_fraction


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path."""
    hook = import_hook.create_hook(
        hook_name=__name__,
        source_init=source_init,
        transform_source=transform_source,
    )
    return hook



"""ideas makes it easy to experiment with alternatives to Python's syntax.

If no source is given, ideas will start an interactive console.
"""

import argparse
from importlib import import_module
import sys

import ideas
from ideas import console
from ideas.session import config


parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=__doc__,
)

parser.add_argument(
    "-v",
    "--version",
    help="Only displays the current version.",
    action="store_true",
)

parser.add_argument(
    "-a",
    "--add_hook",
    nargs=1,
    help="""Execute add_hook() from the specified module.
    An attempt is made to import the specified module from the
    usual entries in sys.path; if it not found, it is then
    imported from ideas.examples.""",
    metavar="MODULE",
)

parser.add_argument(
    "-r",
    "--register_codec",
    nargs=1,
    help="""Execute the named module to register a codec. The specified module
    is either found in the current directory or, if not found,
    from ideas.examples.""",
    metavar="MODULE",
)

parser.add_argument(
    "-s",
    "--show_changes",
    action="store_true",
    help="""Shows the transformed code before it is executed.""",
)

parser.add_argument(
    "source",
    nargs="?",
    help="""Name of the main Python module (path.to.my_program) to be imported.
    """,
)

parser.add_argument(
    "-i", help="""Starts the console after executing a source""", action="store_true"
)


def add_transform(transform):
    """Call the add_hook function for the named module."""
    try:
        module = import_module(transform)
    except (ImportError, ModuleNotFoundError):
        pass
    else:
        try:
            getattr(module, "add_hook")()
        except AttributeError:
            print(f"Module {module} does not contain a function named add_hook")
            return

    path = f"ideas.examples.{transform}"
    try:
        module = import_module(path)
    except ImportError:
        print(f"{path} is not a known transformer.")
    else:
        getattr(module, "add_hook")()


def register_codec(encoding):
    """Executes a module that is expected to register a custom encoding."""
    try:
        import_module(encoding)
    except (ImportError, ModuleNotFoundError):
        pass

    path = f"ideas.examples.{encoding}"
    try:
        import_module(path)
    except ImportError:
        print(f"{path} is not a known codec.")


def main() -> None:
    args = parser.parse_args()
    if args.version:
        print(f"\nideas version {ideas.__version__}")
        return

    config.show_changes = bool(args.show_changes)

    if args.add_hook and args.register_codec:
        print("You can only use one option at a time:")
        print("- either use a source transformation with -a (--add_hook); or")
        print("- register a custom codec with -r (--register_codec),")
        return

    if args.add_hook:
        add_transform(args.add_hook[0])
    if args.register_codec:
        register_codec(args.register_codec[0])

    if args.source is not None:
        if args.source.endswith(".py"):
            args.source = args.source[:-3]
        module = import_module(args.source)
        if sys.flags.interactive or args.i:
            console.start(locals_=module.__dict__)
    else:
        console.start()


main()



from ideas.examples import nfc_normalization
from ideas.import_hook import remove_hook


def test_nfc_normalization():
    hook = nfc_normalization.add_hook()
    from . import n_set  # noqa
    remove_hook(hook)


"""lambda_encoding.py
------------------------

This codec replaces any Python identifier (token) represented by the
single Greek letter 'λ' by the corresponding string 'lambda' which is
the Python keyword.

The source is assumed to be actually encoded in utf-8.
"""

import codecs
import encodings

from . import console

utf8 = encodings.search_function("utf8")


def register_encoding(encoding_name=None, transform_source=None, hook_name=None):
    if encoding_name is None:
        raise TypeError("You must supply a name for your encoding")
    if transform_source is None or not callable(transform_source):
        raise TypeError("You must supply a function as a source transformation.")

    def ideas_decode(input, errors="strict"):
        """Decodes the source as utf8 and use ``transform_source`` to
        perform the required source transformation.
        """
        text, length = utf8.decode(input, errors)
        text = transform_source(text)
        return text, length

    class CustomIncrementalDecoder(encodings.utf_8.IncrementalDecoder):
        def decode(self, input, final=False):
            self.buffer += input
            if final:
                buff = self.buffer
                self.buffer = b""
                text, _ = ideas_decode(buff)
                buff = text.encode("utf-8")
                return super().decode(buff, final=True)
            else:
                return ""

    def search_function(encoding):
        if encoding != encoding_name:
            return None

        return codecs.CodecInfo(
            name=encoding_name,
            encode=utf8.encode,
            decode=ideas_decode,
            incrementalencoder=utf8.incrementalencoder,
            incrementaldecoder=CustomIncrementalDecoder,
            streamreader=encodings.utf_8.StreamReader,
            streamwriter=utf8.streamwriter,
        )

    codecs.register(search_function)
    print(f"{encoding_name} has been registered.")

    if hook_name is not None:
        transform_source.hook_name = hook_name
    console.configure(transform_source=transform_source)



"""utils.py
============

A few utility functions for this project.
"""

import os
import uuid

import token_utils  # to find the path of site-packages

PYTHON = os.path.dirname(os.__file__).lower()
SITE_PACKAGES = os.path.dirname(token_utils.__file__).lower()
IDEAS = os.path.dirname(__file__).lower()
TESTS = os.path.normpath(os.path.join(IDEAS, "..", "tests")).lower()
HOME = os.path.expanduser("~").lower()


def shorten_path(path):
    """Utility function used to reduce the length of the path shown
    to a user, including removing the extension.
    For example, a path for a module in the Python
    standard library might be shown as::

        PYTHON:/module

    whereas a file found in the user's root directory might be shown
    as::

         ~/file
    """
    # On Windows, the filenames are not case-sensitive
    # and the way Python displays filenames may vary.
    # To properly compare, we convert everything to lowercase
    # However, we ensure that the shortened path retains its cases
    ext = os.path.splitext(path)[1]

    path_lower = path.lower()
    path_lower = os.path.splitext(path_lower)[0]

    if path_lower.startswith(PYTHON):
        path = "PYTHON:" + path[len(PYTHON) : -len(ext)]
    elif path_lower.startswith(IDEAS):
        path = "IDEAS:" + path[len(IDEAS) : -len(ext)]
    elif path_lower.startswith(TESTS):
        path = "TESTS:" + path[len(TESTS) : -len(ext)]
    elif path_lower.startswith(SITE_PACKAGES):
        path = "SITE-PACKAGES:" + path[len(SITE_PACKAGES) : -len(ext)]
    elif path_lower.startswith(HOME):
        path = "~" + path[len(HOME) : -len(ext)]
    return path


def print_paths():
    """Prints the values of the path abbreviations used in shorten_path()."""
    print(f"~: {HOME}")
    print(f"PYTHON: {PYTHON}")
    print(f"IDEAS: {IDEAS}")
    print(f"SITE-PACKAGES {SITE_PACKAGES}")
    if os.path.exists(TESTS):
        print(f"TESTS: {TESTS}")


def print_source(source, header="Source"):
    """Prints a maximum of 10 lines of the source code.

    If there is a single line, it is prefixed by ``header: `.
    Otherwise, it is surrounded by dividers.

    ``header`` is usually either ``"Original"`` or ``"New"``
    """
    lines = source.split("\n")
    if len(lines) > 1:
        shortened_source_indicator = "\n..." if len(lines) > 10 else ""
        if len(lines) > 10:
            lines = lines[:10]
        while not lines[-1]:
            lines.pop()
        source = "\n".join(lines[:10]) + shortened_source_indicator
        print(f"\n#========== {header} ====")
        print(source)
        print(f"#=== End of {header} ====\n")
    else:
        print(f"{header}: {source}")


def generate_variable_names():
    """Generator that yields random variable names"""
    while True:
        name = uuid.uuid4()
        yield f"_{name.hex}"


def generate_predictable_names():
    """Generator that yields predictable variable names - useful for testing"""
    index = 0
    while True:
        index += 1
        yield f"_{index}"



"""Python does 'NFKC normalization' of unicode names by default.
As a result, some different unicode names can end up representing
the same object.

This example demonstrates what could happen if we used
'NFC' normalization instead.

Adapted from unnormalized_unicode which was an
original idea from Sergey B. Kirpichev and is found as
anoter example.
"""
import io
import tokenize
import token_utils
import unicodedata
import uuid

from ideas import import_hook

_NAMES_MAP = {}


def transform_source(source, **_kwargs):
    """Transform names that would normally be 'normalized' by
    Python into different and unique variable names.
    """
    # We want to be able to identify when the source has been modified
    # in case session.config.show_changes has been set to True.
    # However, Python's tokenize module cannot untokenize faithfully:
    # some spaces might be messed up.
    # On the other hand, token_utils can untokenize perfectly, recovering
    # the original source. Unfortunately for this case, it works
    # with (already normalized) strings.
    # So, we combine the two here, just to ensure that we can
    # correctly identify later when a source has been modified.

    new_strings = []
    done_normalization = False
    g = tokenize.tokenize(io.BytesIO(source.encode()).readline)
    for token_type, token_string, _, _, _ in g:
        if token_type == tokenize.NAME:
            nkfc_name = unicodedata.normalize("NFKC", token_string)
            nfc_name = unicodedata.normalize("NFC", token_string)
            if nkfc_name != nfc_name:
                if token_string not in _NAMES_MAP:
                    _NAMES_MAP[token_string] = f"{nkfc_name}_{uuid.uuid4().hex!s}"
                new_strings.append(_NAMES_MAP[token_string])
                done_normalization = True
            else:
                new_strings.append(token_string)
    if not done_normalization:
        return source

    new_tokens = []
    tokens = token_utils.tokenize(source)
    for token in tokens:
        if token.is_name():
            token.string = new_strings.pop(0)
        new_tokens.append(token)
    return token_utils.untokenize(new_tokens)


def new_dir(obj=None):
    """Similar to Python's dir, but shows the original name
    entered, not the transformed one.

    Note: the real Python dir() should be available as true_dir().
    """
    import inspect

    if obj is not None:
        names = dir(obj)
    else:
        names = list(inspect.currentframe().f_back.f_locals)
    for k, v in _NAMES_MAP.items():
        names = [name.replace(v, k) for name in names]
    return sorted(names)


def source_init():
    return """true_dir = dir
from ideas.examples.nfc_normalization import new_dir as dir
"""


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    return import_hook.create_hook(
        transform_source=transform_source, source_init=source_init, hook_name=__name__
    )



from ideas.examples import nobreak

def test_for():
    source_for = """
    for i in range(10):
        pass
    %s:
        pass"""

    source = source_for % "nobreak"
    result = nobreak.transform_source(source)
    expected = source_for % "else"

    assert result == expected, "nobreak with for"


def test_while():
    source_while = """
    while True:
        pass
    %s:
        pass"""

    source = source_while % "nobreak"
    result = nobreak.transform_source(source)
    expected = source_while % "else"

    assert result == expected, "nobreak with while"



"""
.. admonition:: Summary

    This example demonstrates the how to use a non-standard file
    extension (``.pyfr``) as an indication that an import hook must
    be used.

    It also demonstrates how to use the ``verbose_finder`` configuration option.

French Python
==============

Imagine you are a French beginner who has learned the basics
of programming using a block-based environment such as Scratch
or Blockly. All the text shown on these blocks was in French,
the only language you know.  You now want to do a transition
to actually writing code in an editor, instead of putting
predefined blocks together. It would be so much easier if
you could use a version of Python where the keywords were in French,
with most of them being identical to what you were using in the
block-based environment.
This is what this import hook example allows one to do.

A more elaborate example is that given by
`AvantPy <https://aroberge.github.io/avantpy/docs/html/>`_.

Let's see it in action:

.. code-block:: none

    > python -m ideas -a french --show
    Ideas Console version 0.0.38. [Python version: 3.9.10]

    ideas> pourchaque lettre dans 'Bonjour':
       ...     afficher(lettre)
       ...
    ===========Transformed============
    for lettre in 'Bonjour':
        print(lettre)

    -----------------------------
    B
    o
    n
    j
    o
    u
    r

    ideas>


Importing .pyfr files
----------------------

Suppose we have the following two files:

.. code-block:: python

   # my_program.py

   print("Wrong one")
   raise ImportError

and

.. code-block:: none


   # my_program.pyfr

   afficher("Bonjour !")


Let's see if we attempt to import ``my_program`` after
setting up the ``french`` import hook and enabling the
verbose finder::


   >>> from ideas.session import config
   >>> config.verbose_finder = True
   >>> from ideas.examples import french
   >>> hook = french.add_hook()
   Looking for files with extensions:  ['.pyfr']
   The following paths will not be included in the search:
      PYTHON: c:\\users\\andre\\appdata\\local\\programs\\python\\python37\\lib
      IDEAS: c:\\users\\andre\\github\\ideas\\ideas
      SITE-PACKAGES: c:\\users\\andre\\github\\ideas\\venv-ideas3.7\\lib\\site-packages
   >>> import my_program
       Searching for ~\\github\\ideas\\my_program.pyfr
       Found: ~\\github\\ideas\\my_program.pyfr

   Bonjour !
   >>> import math
       Searching for ~\\github\\ideas\\math.pyfr
       IdeasMetaFinder did not find math.pyfr

   >>> math.pi
   3.141592653589793
   >>>
"""
from ideas import import_hook
import token_utils

fr_to_py = {
    "Faux": "False",
    "Aucun": "None",
    "Vrai": "True",
    "et": "and",
    "comme": "as",
    "affirmer": "assert",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "interrompre": "break",
    "classe": "class",
    "continuer": "continue",
    "définir": "def",
    "supprimer": "del",
    "sinonsi": "elif",
    "sinon": "else",
    "siexception": "except",
    "finalement": "finally",
    "pourchaque": "for",
    "de": "from",
    "global": "global",
    "si": "if",
    "importer": "import",
    "dans": "in",
    "est": "is",
    "fonction": "lambda",
    "nonlocal": "nonlocal",
    "pas": "not",
    "ou": "or",
    "passer": "pass",
    "lever": "raise",
    "retourner": "return",
    "essayer": "try",
    "tantque": "while",
    "avec": "with",
    "céder": "yield",
    # a few builtins useful for beginners
    "demander": "input",
    "afficher": "print",
    "intervalle": "range",
    "quitter": "exit",  # useful for console
}


def transform_source(source, **_kwargs):
    """A simple replacement of 'French Python keyword' by their normal
    English version.
    """
    new_tokens = []
    for token in token_utils.tokenize(source):
        if token.string in fr_to_py:
            token.string = fr_to_py[token.string]
        new_tokens.append(token)

    new_source = token_utils.untokenize(new_tokens)
    return new_source


def add_hook(**_kwargs):
    """Creates and adds the import hook in sys.meta_path.
    Uses a custom extension for the exception hook."""
    hook = import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,
        extensions=[".pyfr"],
    )
    return hook



"""decimal_math_d.py
------------------------

This replaces any explicit float followed by ``D``, by a Decimal.
"""


from ideas import import_hook
import token_utils


def source_init():
    """Adds required import"""
    return "from decimal import Decimal\n"


def transform_source(source, **_kwargs):
    """Simple transformation: replaces any explicit float followed by ``D``
    by a Decimal.
    """
    tokens = token_utils.tokenize(source)
    for first, second in zip(tokens, tokens[1:]):
        if first.is_number() and "." in first.string and second == "D":
            first.string = f"Decimal('{first.string}')"
            second.string = ""

    return token_utils.untokenize(tokens)


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        source_init=source_init,
        transform_source=transform_source,
    )
    return hook



"""import_hook.py
------------------

This module contains the core functions required to create an import hook.
"""

import ast
import os
import sys
import traceback

from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location, decode_source
from types import CodeType, ModuleType
from typing import Callable, Dict, Sequence, Optional, Any

from . import console
from .session import config
from .utils import shorten_path, PYTHON, IDEAS, SITE_PACKAGES, print_source


def finder_inform(text):
    """Print some informative text when verbose finder is set"""
    if config.verbose_finder:
        print(text)


class IdeasMetaFinder(MetaPathFinder):  # pylint: disable=R0902
    """A custom finder to locate modules. The main reason for this code
    is to ensure that our custom loader, which does the code transformations,
    is used."""

    def __init__(
        self,
        callback_params=None,
        create_module=None,
        excluded_paths=None,
        exec_=None,
        extensions=None,
        hook_name=None,
        module_class=None,
        source_init=None,
        transform_ast=None,
        transform_bytecode=None,
        transform_source=None,
    ):  # pylint: disable=R0913
        self.callback_params = callback_params
        self.custom_create_module = create_module
        self.excluded_paths = excluded_paths
        self.exec_ = exec_
        self.extensions = extensions
        self.hook_name = hook_name
        self.module_class = module_class
        self.source_init = source_init
        self.transform_ast = transform_ast
        self.transform_bytecode = transform_bytecode
        self.transform_source = transform_source

    def __repr__(self):
        if self.hook_name is None:
            return "<IdeasMetaFinder object>"
        return f"<IdeasMetaFinder object for {self.hook_name}>"

    def find_spec(self, fullname, path, target=None):  # pylint: disable=W0613
        """finds the appropriate properties (spec) of a module, and sets
        its loader."""
        if not path:
            path = [os.getcwd()]

        if "." in fullname:
            name = fullname.split(".")[-1]
        else:
            name = fullname

        for entry in path:
            skip = False
            for sub_path in self.excluded_paths:
                if sub_path in entry.lower():
                    skip = True
                    if config.verbose_finder:
                        print("    Skipping over:", shorten_path(entry))
                    break
            if skip:
                continue

            for extension in self.extensions:
                if not extension.startswith("."):  # be forgiving ...
                    extension = "." + extension
                filename = os.path.join(entry, name + extension)

                finder_inform(f"    Searching for {shorten_path(filename)}{extension}")
                if os.path.exists(filename):
                    finder_inform(f"    Found: {shorten_path(filename) + extension}\n")
                    break
                finder_inform(
                    "    IdeasMetaFinder did not find"
                    + f"{shorten_path(fullname)}{extension}\n",
                )
            else:
                continue

            return spec_from_file_location(
                fullname,
                filename,
                loader=IdeasLoader(
                    filename,
                    callback_params=self.callback_params,
                    create_module=self.custom_create_module,
                    exec_=self.exec_,
                    module_class=self.module_class,
                    source_init=self.source_init,
                    transform_ast=self.transform_ast,
                    transform_bytecode=self.transform_bytecode,
                    transform_source=self.transform_source,
                ),
            )
        return None  # we don't know how to import this


class IdeasLoader(Loader):  # pylint: disable=R0902
    """A custom loader which will transform the source prior to its execution"""

    def __init__(
        self,
        filename,
        callback_params=None,
        create_module=None,
        exec_=None,
        module_class=None,
        source_init=None,
        transform_ast=None,
        transform_bytecode=None,
        transform_source=None,
    ):  # pylint: disable=R0913
        self.filename = filename
        self.exec_ = exec_
        self.callback_params = callback_params
        self.custom_create_module = create_module
        self.module_class = module_class
        self.source_init = source_init
        self.transform_ast = transform_ast
        self.transform_bytecode = transform_bytecode
        self.transform_source = transform_source
        # Identify the main script assuming that it has been called from
        # the command line using something like
        # python -m ideas main_script[.py] -some_flag
        self.main_name = None
        if len(sys.argv) > 1 and not sys.argv[1].startswith("-"):
            self.main_name = sys.argv[1].rstrip(".py")

    def create_module(self, spec):
        """Potential replacement for the default create_module method."""
        # Note: I do not have an example of custom module creation yet.
        if self.custom_create_module is not None:
            return self.custom_create_module(spec, callback_params=self.callback_params)
        return None  # use default module creation semantics

    def exec_module(self, module):
        """Import the source code, transform it before executing it so that
        it is known to Python.
        """
        # The main script should be the very first one imported.
        # To avoid potential named conflicts, we ensure that the potential
        # identification is only done once.
        if module.__name__ is not None and module.__name__ == self.main_name:
            module.__name__ = "__main__"
        self.main_name = None

        if self.module_class is not None:
            module.__class__ = self.module_class  # pylint: disable=E0243

        with open(self.filename, mode="r+b") as file:
            source = decode_source(file.read())
        original_source = source

        if self.transform_source is not None:
            try:
                source = self.transform_source(
                    source,
                    filename=self.filename,
                    module=module,
                    callback_params=self.callback_params,
                )
            except Exception as exc:
                print("Error when executing transform_source:")
                print("   ", traceback.format_exception_only(type(exc), exc)[0].strip())
                return

        if config.show_changes and original_source != source:
            print_source(original_source, header="Original")
            print_source(source, header="New")

        if self.source_init is not None:
            source = self.source_init() + source

        try:
            tree = ast.parse(source, self.filename)
        except Exception:
            print("Exception raised while parsing source.")
            raise

        if self.transform_ast is not None:
            tree = self.transform_ast(tree)

        try:
            code_object = compile(tree, self.filename, "exec")
        except Exception:
            print("Exception raised while compiling tree.")
            raise

        if self.transform_bytecode is not None:
            code_object = self.transform_bytecode(code_object)

        if self.exec_ is not None:
            self.exec_(
                code_object,
                filename=self.filename,
                globals_=module.__dict__,
                module=module,
                callback_params=self.callback_params,
            )
        else:
            try:
                exec(code_object, module.__dict__)  # pylint: disable=W0122
            except Exception:
                print("Exception raised while executing code object.")
                raise


def create_hook(
    callback_params: Optional[Dict[str, Any]] = None,
    create_module: Optional[Callable[..., ModuleType]] = None,
    console_dict: Optional[Dict[str, Any]] = None,
    exec_: Optional[Callable[..., None]] = None,
    extensions: Optional[Sequence[str]] = None,
    first: bool = True,
    hook_name: Optional[str] = None,
    ipython_ast_node_transformer: Optional[ast.NodeTransformer] = None,
    module_class: Optional[type] = None,
    source_init: Optional[Callable[[], str]] = None,
    transform_ast: Optional[ast.NodeTransformer] = None,
    transform_bytecode: Optional[Callable[[CodeType], CodeType]] = None,
    transform_source: Optional[Callable[[str], str]] = None,
) -> IdeasMetaFinder:  # pylint: disable=R0913,R0914
    """Function to facilitate the creation of an import hook.

    Each of the following parameter is optional; most of these are
    never needed except in some unusual import hooks.

    Usually, at least one of ``transform_ast``, ``transform_bytecode``,
    and ``transform_source`` should be specified.

    * ``callback_params``: a dict containing keyword parameters
      to be passed back to the ``transform_source`` function.
    * ``create_module``: a custom function to create a module object
      instead of using Python's default.
    * ``console_dict``: a dict object used as 'locals' with the Ideas console,
      instead of its usual default.
    * ``exec_``: a custom method used to execute the source code inside
      a module's dict.
    * ``extensions``: a list of file extensions, other than the usual `.py`, etc.,
      used to identify modules containing source code.
    * ``first``: if ``True``, the custom hook will be used as the first
      location in ``sys.meta_path``, to look for source files.
    * ``hook_name``: used to give a more readable ``repr`` to the hook created.
    * ``ipython_ast_node_transformer``: used to do AST transformations in an
      IPython/Jupyter environment. It should be a class derived from
      ``ast.NodeTransformer`` and return a ``node``.
    * ``module_class``: custom class to use for the module created instead of
      the default one assigned by Python.
    * ``source_init``: custom code to be executed before any code from
      a user is executed. For example, if one creates an import hook that
      treats every ``float`` as a ``Decimal`` object, this custom code
      could be::

          from decimal import Decimal

    * ``transform_ast``: used to do AST transformations in a Python
      environment (excluding IPython/Jupyter).  It should be a class
      derived from ``ast.NodeTransformer``, eventually returning a
      tree object.
    * ``transform_bytecode``: used to mutate a code object.
    * ``transform_source``: used to transform some source code prior
      to execution.
    """
    if extensions is None:
        extensions = [".py"]

    try:
        ipython_shell = get_ipython()  # noqa
    except NameError:
        ipython_shell = None

    if source_init is not None and source_init().strip() and ipython_shell is not None:
        print("   The following initializing code from ideas is included:\n")
        print(source_init().strip())
        lines = [line for line in source_init().splitlines() if line.strip()]
        for line in lines:
            ipython_shell.ex(line)

    excluded_paths = [PYTHON, IDEAS, SITE_PACKAGES]

    if config.verbose_finder:
        print("Looking for files with extensions: ", extensions)
        print("The following paths will not be included in the search:")
        for sub_path in excluded_paths:
            print("  ", shorten_path(sub_path), sub_path)

    hook = IdeasMetaFinder(
        callback_params=callback_params,
        create_module=create_module,
        excluded_paths=excluded_paths,
        exec_=exec_,
        extensions=extensions,
        hook_name=hook_name,
        module_class=module_class,
        source_init=source_init,
        transform_ast=transform_ast,
        transform_bytecode=transform_bytecode,
        transform_source=transform_source,
    )

    if first:
        sys.meta_path.insert(0, hook)
    else:
        sys.meta_path.append(hook)

    for obj in [transform_source, transform_ast, transform_bytecode, source_init]:
        if obj is not None and isinstance(hook_name, str):
            obj.hook_name = hook_name

    console.configure(
        callback_params=callback_params,
        console_dict=console_dict,
        source_init=source_init,
        transform_ast=transform_ast,
        transform_bytecode=transform_bytecode,
        transform_source=transform_source,
    )
    if transform_source is not None and ipython_shell is not None:
        ipython_source_transformer = make_ipython_source_transformer(transform_source)
        ipython_shell.input_transformers_cleanup.append(ipython_source_transformer)

    if ipython_ast_node_transformer is not None and ipython_shell is not None:
        wrapped_ipython_ast_node_transformer = make_ipython_ast_node_transformer(
            ipython_ast_node_transformer
        )
        ipython_shell.ast_transformers.append(wrapped_ipython_ast_node_transformer())

    return hook


def make_ipython_source_transformer(transform_source):
    """Takes a source transform and makes returns an IPython compatible
    source transformer.
    """
    # This is done as during the cleanup phase
    # (``ipython_shell.input_transformers_cleanup``), as opposed to the
    # post phase (``ipython_shell.input_transformers_post``) so that
    # transformations that work on code blocks (such as ``repeat``)
    # can work properly.
    def ipython_source_transformer(lines):  # noqa
        # In IPython, the source transformation operates on a list of lines
        original_source = "".join(lines)
        source = transform_source(original_source)
        if config.show_changes and source != original_source:
            config.print_transformed(source, header="New: ")
        lines = source.splitlines(keepends=True)
        return lines

    return ipython_source_transformer


def make_ipython_ast_node_transformer(ipython_ast_node_transformer):
    """Takes an AST transformer designed to work with IPython,
    and wraps it to add a warning in case the user would like to
    see how the code is actually transformed, since this is not
    possible when using IPython.
    """

    def wrapped_ipython_ast_node_transformer():
        if config.show_changes:
            print(
                "Cannot show the changed source for AST transform in IPython/Jupyter."
            )
            config.show_changes = False
        return ipython_ast_node_transformer

    return wrapped_ipython_ast_node_transformer


def remove_hook(hook):
    """Function used to remove a previously imported hook inserted in sys.meta_path"""
    for index, meta_finder in enumerate(sys.meta_path):
        if meta_finder == hook:
            del sys.meta_path[index]
            break
    else:
        print("Import hook not found in remove_hook.")
        return



from ideas.examples import french
from ideas.import_hook import remove_hook


def test_french():
    hook = french.add_hook()
    from . import mon_programme
    assert mon_programme.carré(4) == 16, "The square of 4 is 16"  # noqa
    remove_hook(hook)



"""
.. admonition:: Summary

    This module enables someone to use ``nobreak`` as a keyword
    equivalent to ``else`` in ``for`` and ``while`` loops.

``nobreak`` as a keyword
========================

Python's ``for`` and ``while`` loop include an ``else`` clause
whose meaning is not immediately obvious::

    while condition:
        # some
        # code
        # here
    else:
        # will be executed only if no
        # break statement occurred above

When I first understood this, I thought *wouldn't it be nice if, instead
of using* ``else:``, *one could write something like* ``if not break:`` which
uses only existing Python keywords.
For this example, I decided instead that a suggestion made by Raymond Hettinger
to have ``nobreak`` as a keyword made the most sense.
It can be used instead of ``else`` in the above example::

    while condition:
        # some
        # code
        # here
    nobreak:
        # will be executed only if no
        # break statement occurred above


``nobreak`` instead of ``else`` in ``if/else``
-------------------------------------------------------

The ``else`` keyword has a very different meaning when used as part
of an ``if`` statement.  In this situation, ``nobreak``, or its
translation in some other language would make no sense.

As a result, if one attempts to write the following::

    if condition:
        # some
        # code
        # here
    nobreak:
        # more code

``nobreak`` will **not** be replaced by ``else`` and
the code will raise a ``SyntaxError``.


What about try/except?
-----------------------

The ``else`` keyword can also be used in a ``try/except/else/finally`` block.
From `Python's documentation <https://docs.python.org/3/reference/compound_stmts.html#the-try-statement>`_:

   *The optional else clause is executed if the control flow leaves the try suite,*
   **no exception was raised**,
   *and no* ``return``, ``continue``, *or* ``break`` *statement was executed.*

Since multiple causes can prevent the ``else`` clause from being executed,
it makes little sense in this case to use a different keyword such as
``nobreak``, that would point to a specific cause which would likely be wrong.

Useful function from ``token_utils``
-------------------------------------

The code for ``nobreak`` makes use of the ``get_first`` function from
``token_utils``. Here's some useful information about it.

.. code-block::

    >>> import token_utils
    >>> help(token_utils.get_first)
    Help on function get_first in module token_utils:

    get_first(tokens, exclude_comment=True)
        Given a list of tokens, find the first token which is not a space token
        (such as a ``NEWLINE``, ``INDENT``, ``DEDENT``, etc.) and,
        by default, also not a ``COMMMENT``.

        ``COMMMENT`` tokens can be included by setting ``exclude_comment`` to ``False``.

        Returns ``None`` if none is found.

"""
from ideas import import_hook
import token_utils


def transform_source(source, **_kwargs):
    """``nobreak`` is replaced by ``else`` only if it is the first
    non-space token on a line and if its indentation matches
    that of a ``for`` or ``while`` block.
    """
    indentations = {}
    lines = token_utils.get_lines(source)
    new_tokens = []
    # The following is not a proper parser, but it should work
    # well enough in most cases, for well-formatted code.
    for line in lines:
        first = token_utils.get_first(line)
        if first is None:
            new_tokens.extend(line)
            continue
        if first == "nobreak":
            if first.start_col in indentations:
                if indentations[first.start_col] in ["for", "while"]:
                    first.string = "else"
                    del indentations[first.start_col]
        indentations[first.start_col] = first.string
        new_tokens.extend(line)

    return token_utils.untokenize(new_tokens)


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    return import_hook.create_hook(
        transform_source=transform_source,
        hook_name=__name__,
    )



"""confused_math_bc.py
------------------------

Import hook that performs bytecode changes,
swapping BINARY_ADD and BINARY_MULTIPLY.

Note: If code contains something like a + b, or a * b, where
a and b are integer literals (e.g. 2 + 4), the operations are
done prior to the creation of a code object and thus are
not captured by this transformation.
"""
import dis
import sys
from types import CodeType
from ideas import import_hook

# The actual bytecode used can change with different Python version.
# Instead of attempting to hard-code some actual bytecode values
# we extract the actual value used in the current Python version.
if sys.version_info.minor < 11:
    ADD = dis.opmap["BINARY_ADD"]
    MUL = dis.opmap["BINARY_MULTIPLY"]
    BINARY_OP = -1
    binop_arg_add = binop_arg_mul = -1
else:
    ADD = MUL = -1
    BINARY_OP = dis.opmap["BINARY_OP"]
    for index, op_nb in enumerate(dis._nb_ops):  # noqa
        if op_nb[1] == "+":
            binop_arg_add = index
        elif op_nb[1] == "*":
            binop_arg_mul = index


def swap_add_mul(bytecode):
    """Interchange BINARY_ADD and BINARY_MULTIPLY"""
    new_bc = []
    check_arg = -1
    for i, b in enumerate(bytecode):
        if b == ADD and i % 2 == 0:
            new_bc.append(MUL)
        elif b == MUL and i % 2 == 0:
            new_bc.append(ADD)
        elif b == BINARY_OP and i % 2 == 0:
            new_bc.append(b)
            check_arg = i + 1
            continue
        elif i == check_arg:
            if b == binop_arg_add:
                new_bc.append(binop_arg_mul)
            elif b == binop_arg_mul:
                new_bc.append(binop_arg_add)
            else:
                new_bc.append(b)
        else:
            new_bc.append(b)
    return bytes(new_bc)


def create_new_co(code_object):
    """Recursively creates new code objects from old ones, swapping
    BINARY_ADD and BINARY_MULTIPLY.
    """
    new_code = swap_add_mul(code_object.co_code)
    new_const = []
    for c in code_object.co_consts:
        if isinstance(c, CodeType):
            new_const.append(create_new_co(c))
        else:
            new_const.append(c)

    if sys.version_info.minor in (6, 7):
        new_code_object = CodeType(
            code_object.co_argcount,
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_object.co_flags,
            new_code,
            tuple(new_const),
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_firstlineno,
            code_object.co_lnotab,
            code_object.co_freevars,
            code_object.co_cellvars,
        )
    elif sys.version_info.minor < 11:
        new_code_object = CodeType(
            code_object.co_argcount,
            code_object.co_posonlyargcount,  # new in Python 3.8
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_object.co_flags,
            new_code,
            tuple(new_const),
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_firstlineno,
            code_object.co_lnotab,
            code_object.co_freevars,
            code_object.co_cellvars,
        )
    else:
        new_code_object = CodeType(  # noqa
            code_object.co_argcount,
            code_object.co_posonlyargcount,
            code_object.co_kwonlyargcount,
            code_object.co_nlocals,
            code_object.co_stacksize,
            code_object.co_flags,
            new_code,
            tuple(new_const),
            code_object.co_names,
            code_object.co_varnames,
            code_object.co_filename,
            code_object.co_name,
            code_object.co_qualname,  # new in 3.11
            code_object.co_firstlineno,
            code_object.co_lnotab,
            code_object.co_exceptiontable,  # new in 3.11
            code_object.co_freevars,
            code_object.co_cellvars,
        )
    return new_code_object


def transform_bytecode(code_object, **_kwargs):
    """Transforms the code object into a new code object"""
    new_code_object = create_new_co(code_object)
    return new_code_object


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        transform_bytecode=transform_bytecode,
    )
    return hook



"""
.. admonition:: Summary

    - Demonstrates how to use an import hook that does an AST transformation

Fractional math (AST)
==========================


Consider the following standard Python code::

    >>> x = 1/10
    >>> for i in range(11):
    ...    print(i * x)
    ...
    0.0
    0.1
    0.2
    0.30000000000000004
    0.4
    0.5
    0.6000000000000001
    0.7000000000000001
    0.8
    0.9
    1.0

This is quite surprising for beginners, not familiar with the
limitations of representing floating point numbers.

However, we can "fix" this using an import hook that performs
and Abstract Syntax Tree (AST) transformation::


    > python -m ideas -a fractions_ast
       The following initializing code from ideas is included:

    from fractions import Fraction

    Ideas Console version 0.0.38. [Python version: 3.10.2]

    ideas> x = 1/10

    ideas> for i in range(11):
       ...     print(i * x)
       ...
    0
    1/10
    1/5
    3/10
    2/5
    1/2
    3/5
    7/10
    4/5
    9/10
    1

    ideas> from ideas.session import config

    ideas> config.show_changes = True

    ideas> x = 1/10
    new: x = Fraction(1, 10)

    ideas>
"""
import ast

from ideas import import_hook


class FractionWrapper(ast.NodeTransformer):
    """Wrap all int divisions in a call to Fractions.

    Adapted from https://github.com/diofant/diofant"""

    def visit_BinOp(self, node):
        def is_integer(x):
            if isinstance(x, ast.Num) and isinstance(x.n, int):
                return True
            elif isinstance(x, ast.UnaryOp) and isinstance(x.op, (ast.USub, ast.UAdd)):
                return is_integer(x.operand)
            elif isinstance(x, ast.BinOp) and isinstance(
                x.op, (ast.Add, ast.Sub, ast.Mult, ast.Pow)
            ):
                return is_integer(x.left) and is_integer(x.right)
            else:
                return False

        if isinstance(node.op, ast.Div) and all(
            is_integer(_) for _ in [node.left, node.right]
        ):
            return ast.Call(
                func=ast.Name(id="Fraction", ctx=ast.Load()),
                args=[node.left, node.right],
                keywords=[],
                starargs=None,
                kwargs=None,
            )
        # IPython requires transformations that return a single node
        # whereas ideas works with entire trees.
        try:
            get_python()  # noqa
            return node
        except NameError:
            return self.generic_visit(node)


ipython_ast_node_transformer = FractionWrapper()


def transform_ast(tree, **_kwargs):
    """Transforms the Abstract Syntax Tree or a single node"""
    tree_or_node = FractionWrapper().visit(tree)
    ast.fix_missing_locations(tree_or_node)
    return tree_or_node


def source_init():
    """Adds required import so that ``Fraction`` is a known object."""
    import_fraction = "from fractions import Fraction\n"
    return import_fraction


def add_hook(**_kwargs):
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        source_init=source_init,
        transform_ast=transform_ast,
        ipython_ast_node_transformer=ipython_ast_node_transformer,
    )
    return hook



from ideas.examples import unnormalized_unicode
from ideas.import_hook import remove_hook


def test_unnormalized_unicode():
    hook = unnormalized_unicode.add_hook()
    from . import n_set  # noqa
    remove_hook(hook)


from ideas.examples import confused_math_bc
from ideas.import_hook import remove_hook


def test_simple_case():
    hook = confused_math_bc.add_hook()
    from . import confused_math  # noqa
    remove_hook(hook)



from ideas.examples import switch
from ideas.import_hook import remove_hook


def test_transform():
    source = """
        switch EXPR:
            case EXPR_1:
                SUITE
            case EXPR_2:
                SUITE
            case in (EXPR_3, EXPR_4, ...):
                SUITE
            else:
                SUITE
        other_code"""

    expected = """
        _1 = EXPR
        if _1 == EXPR_1:
                SUITE
        elif _1 == EXPR_2:
                SUITE
        elif _1 in (EXPR_3, EXPR_4, ...):
                SUITE
        else:
                SUITE
        del _1
        other_code"""

    result = switch.transform_source(source, callback_params={"predictable_names": True})
    assert result == expected, "Switch conversion test"


def test_switch1():
    hook = switch.add_hook()
    from . import switch1  # noqa
    remove_hook(hook)
